/*
 * mautrix-telegram - A Matrix-Telegram puppeting bridge
 * Copyright (C) 2019 Tulir Asokan
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
form > div {
	display: none;
}

form[data-status="request"] > div.status-request,
form[data-status="code"] > div.status-code,
form[data-status="password"] > div.status-password {
	display: initial;
}

.container {
	margin-top: 3rem;
	max-width: 60rem;
}

.error, .message {
	border-radius: .25rem;
	padding: .5rem 1rem;
	border: 1px solid transparent;
	margin: .5rem 0;
}

.error {
	border-color: #f5c6cb;
	background-color: #f8d7da;
	color: #721c24;
}

.message {
	border-color: #c3e6cb;
	background-color: #d4edda;
	color: #155724;
}

[type="checkbox"], [type="radio"] {
	position: absolute;
	opacity: 0;
}

[type="checkbox"] + label, [type="radio"] + label {
	position: relative;
	padding-left: 2.5rem;
	cursor: pointer;
	display: inline-block;
}

[type="checkbox"] + label:before, [type="radio"] + label:before {
	content: '';
	position: absolute;
	left: 0;
	top: 0.4rem;
	width: 1.8rem;
	height: 1.8rem;
	border: 0.1rem solid #d1d1d1;
}

[type="radio"] + label:before, [type="radio"] + label:after {
	border-radius: 50%;
}

[type="checkbox"]:checked + label:after,
[type="radio"]:checked + label:after {
	content: '';
	width: 0.8rem;
	height: 0.8rem;
	background: #9b4dca;
	position: absolute;
	top: 0.9rem;
	left: 0.5rem;
}

[type="radio"]:disabled + label:before, [type="checkbox"]:disabled + label:before {
	background-color: #d1d1d1;
}

[type="radio"]:disabled + label, [type="checkbox"]:disabled + label {
	color: #d1d1d1;
}

[type="radio"]:disabled:checked + label:after, [type="checkbox"]:disabled:checked + label:after {
	background: #606c76;
}
